use crate::Atom;
use crate::error::Error;
use crate::error::ExpectedKind;
use crate::error::io_eof;

use crate::hex;

use crate::CompoundClass;
use crate::Domain;
use crate::DomainDecode;
use crate::Map;
use crate::Reader;
use crate::Record;
use crate::Set;
use crate::Value;
use crate::boundary as B;
use crate::reader::ReaderResult;
use crate::source::BinarySource;

use num::bigint::BigInt;

use std::io;
use std::iter::FromIterator;
use std::marker::PhantomData;

pub struct TextReader<'de, 'src, S: BinarySource<'de>> {
    pub source: &'src mut S,
    phantom: PhantomData<&'de ()>,
}

impl<'de, 'src, S: BinarySource<'de>> TextReader<'de, 'src, S>
{
    pub fn new(source: &'src mut S) -> Self {
        TextReader {
            source,
            phantom: PhantomData,
        }
    }

    fn syntax_error(&mut self, message: &str) -> io::Error {
        self.source.syntax_error(message)
    }

    fn peek(&mut self) -> io::Result<Option<u8>> {
        self.source.peek()
    }

    #[inline(always)]
    fn peek_noeof(&mut self) -> io::Result<u8> {
        self.source.peek()?.ok_or_else(io_eof)
    }

    fn skip(&mut self) -> io::Result<()> {
        self.source.skip()
    }

    fn next_byte(&mut self) -> io::Result<u8> {
        let b = self.peek_noeof()?;
        self.source.skip()?;
        Ok(b)
    }

    fn skip_whitespace(&mut self) {
        // Deliberately swallows errors.
        while let Ok(Some(c)) = self.peek() {
            match c {
                b' ' | b'\t' | b'\r' | b'\n' | b',' => {
                    let _ = self.skip();
                    ()
                }
                _ => break,
            }
        }
    }

    fn decode_utf8(&mut self, bs: Vec<u8>) -> io::Result<String> {
        String::from_utf8(bs).map_err(|_| self.syntax_error("Invalid UTF-8"))
    }

    fn comment_line(&mut self) -> io::Result<String> {
        let mut bs = Vec::new();
        loop {
            let b = self.peek_noeof()?;
            self.skip()?;
            match b {
                b'\r' | b'\n' => return Ok(self.decode_utf8(bs)?),
                _ => bs.push(b),
            }
        }
    }

    fn read_intpart(&mut self, mut bs: Vec<u8>, c: u8) -> io::Result<Atom<'de>> {
        match c {
            b'0' => {
                bs.push(c);
                self.read_fracexp(bs)
            }
            _ => {
                self.read_digit1(&mut bs, c)?;
                self.read_fracexp(bs)
            }
        }
    }

    fn read_fracexp(&mut self, mut bs: Vec<u8>) -> io::Result<Atom<'de>> {
        let mut is_float = false;
        match self.peek_noeof() {
            Ok(b'.') => {
                is_float = true;
                bs.push(self.next_byte()?);
                let c = self.next_byte()?;
                self.read_digit1(&mut bs, c)?;
            }
            _ => ()
        }
        match self.peek_noeof() {
            Ok(b'e') | Ok(b'E') => {
                bs.push(self.next_byte()?);
                self.read_sign_and_exp(bs)
            }
            _ => self.finish_number(bs, is_float)
        }
    }

    fn read_sign_and_exp(&mut self, mut bs: Vec<u8>) -> io::Result<Atom<'de>> {
        match self.peek_noeof()? {
            b'+' | b'-' => bs.push(self.next_byte()?),
            _ => (),
        }
        let c = self.next_byte()?;
        self.read_digit1(&mut bs, c)?;
        self.finish_number(bs, true)
    }

    fn finish_number(&mut self, bs: Vec<u8>, is_float: bool) -> io::Result<Atom<'de>> {
        let s = self.decode_utf8(bs)?;
        if is_float {
            match self.peek_noeof() {
                Ok(b'f') | Ok(b'F') => {
                    self.skip()?;
                    Ok(Atom::Float(s.parse::<f32>().map_err(
                        |_| self.syntax_error(&format!(
                            "Invalid single-precision number: {:?}", s)))?))
                }
                _ =>
                    Ok(Atom::Double(s.parse::<f64>().map_err(
                        |_| self.syntax_error(&format!(
                            "Invalid double-precision number: {:?}", s)))?))
            }
        } else {
            Ok(Atom::SignedInteger(s.parse::<BigInt>().map_err(
                |_| self.syntax_error(&format!(
                    "Invalid signed-integer number: {:?}", s)))?))
        }
    }

    fn read_digit1(&mut self, bs: &mut Vec<u8>, c: u8) -> io::Result<()>
    {
        if !(c as char).is_digit(10) {
            return Err(self.syntax_error("Incomplete number"));
        }
        bs.push(c);
        while let Ok(Some(c)) = self.peek() {
            if !(c as char).is_digit(10) {
                break;
            }
            bs.push(self.next_byte()?);
        }
        Ok(())
    }

    fn read_stringlike<X, H, R>(
        &mut self,
        mut seed: R,
        xform_item: X,
        terminator: u8,
        hexescape: u8,
        hexescaper: H,
    ) -> io::Result<R>
    where
        X: Fn(&mut Self, &mut R, u8) -> io::Result<()>,
        H: Fn(&mut Self, &mut R) -> io::Result<()>,
    {
        loop {
            match self.next_byte()? {
                c if c == terminator => return Ok(seed),
                b'\\' => match self.next_byte()? {
                    c if c == hexescape => hexescaper(self, &mut seed)?,
                    c if c == terminator || c == b'\\' || c == b'/' => xform_item(self, &mut seed, c)?,
                    b'b' => xform_item(self, &mut seed, b'\x08')?,
                    b'f' => xform_item(self, &mut seed, b'\x0c')?,
                    b'n' => xform_item(self, &mut seed, b'\x0a')?,
                    b'r' => xform_item(self, &mut seed, b'\x0d')?,
                    b't' => xform_item(self, &mut seed, b'\x09')?,
                    _ => return Err(self.syntax_error("Invalid escape code")),
                },
                c => xform_item(self, &mut seed, c)?,
            }
        }
    }

    fn hexnum(&mut self, count: usize) -> io::Result<u32> {
        let mut v: u32 = 0;
        for _ in 0 .. count {
            let c = self.next_byte()?;
            match (c as char).to_digit(16) {
                Some(d) =>
                    v = v << 4 | d,
                None =>
                    return Err(self.syntax_error("Bad hex escape")),
            }
        }
        Ok(v)
    }

    fn append_codepoint(&mut self, bs: &mut Vec<u8>, n: u32) -> io::Result<()> {
        let c = char::from_u32(n).ok_or_else(|| self.syntax_error("Bad code point"))?;
        let mut buf = [0; 4];
        let _ = c.encode_utf8(&mut buf);
        bs.extend(&buf[0 .. c.len_utf8()]);
        Ok(())
    }

    fn read_string(&mut self, delimiter: u8) -> io::Result<String> {
        let raw = self.read_stringlike(
            Vec::new(),
            |_r, bs, c| Ok(bs.push(c)),
            delimiter,
            b'u',
            |r, bs| {
                let n1 = r.hexnum(4)?;
                if (0xd800 ..= 0xdbff).contains(&n1) {
                    let mut ok = true;
                    ok = ok && r.next_byte()? == b'\\';
                    ok = ok && r.next_byte()? == b'u';
                    if !ok {
                        Err(r.syntax_error("Missing second half of surrogate pair"))
                    } else {
                        let n2 = r.hexnum(4)?;
                        if (0xdc00 ..= 0xdfff).contains(&n2) {
                            let n = ((n1 - 0xd800) << 10) + (n2 - 0xdc00) + 0x10000;
                            r.append_codepoint(bs, n)
                        } else {
                            Err(r.syntax_error("Bad second half of surrogate pair"))
                        }
                    }
                } else {
                    r.append_codepoint(bs, n1)
                }
            })?;
        self.decode_utf8(raw)
    }

    fn read_literal_binary(&mut self) -> io::Result<Atom<'de>> {
        Ok(Atom::Bytes(&self.read_stringlike(
            Vec::new(),
            |_r, bs, b| Ok(bs.push(b)),
            b'"',
            b'x',
            |r, bs| Ok(bs.push(r.hexnum(2)? as u8)))?[..]))
    }

    fn read_hex_binary(&mut self) -> io::Result<Atom<'de>> {
        let mut s = String::new();
        loop {
            self.skip_whitespace();
            let c1 = self.next_byte()? as char;
            if c1 == '"' {
                let bs = hex::HexParser::Strict.decode(&s).unwrap();
                return Ok(Atom::Bytes(&bs[..]));
            }
            let c2 = self.next_byte()? as char;
            if !(c1.is_digit(16) && c2.is_digit(16)) {
                return Err(self.syntax_error("Invalid hex binary"));
            }
            s.push(c1);
            s.push(c2);
        }
    }

    fn read_base64_binary(&mut self) -> io::Result<Atom<'de>> {
        let mut bs = Vec::new();
        loop {
            self.skip_whitespace();
            let mut c = self.next_byte()?;
            if c == b']' {
                let bs = base64::decode_config(&self.decode_utf8(bs)?, base64::STANDARD_NO_PAD)
                    .map_err(|_| self.syntax_error("Invalid base64 character"))?;
                return Ok(Atom::Bytes(&bs[..]));
            }
            if c == b'-' { c = b'+'; }
            if c == b'_' { c = b'/'; }
            if c == b'=' { continue; }
            bs.push(c);
        }
    }

    fn read_raw_symbol(&mut self, mut bs: Vec<u8>) -> io::Result<Atom<'de>> {
        loop {
            let c = match self.peek()? {
                None => b' ',
                Some(c) if (c as char).is_whitespace() => b' ',
                Some(c) => c
            };
            match c {
                b'(' | b')' | b'{' | b'}' | b'[' | b']' | b'<' | b'>' |
                b'"' | b';' | b',' | b'@' | b'#' | b':' | b'|' | b' ' =>
                    return Ok(Atom::Symbol(&self.decode_utf8(bs)?)),
                c => {
                    self.skip()?;
                    bs.push(c)
                }
            }
        }
    }
}

impl<'de, 'src, S: BinarySource<'de>> Reader<'de> for TextReader<'de, 'src, S>
{
    fn peek_class(&mut self) -> io::Result<Option<crate::reader::NextToken>> {
        self.skip_whitespace();
        let c = match self.peek()? {
            None => return Ok(None),
            Some(c) => c,
        };
XX        Ok(Some(match c {
            b'-' => {
                self.skip()?;
                let c1 = self.next_byte()?;
                self.read_intpart(vec![b'-'], c1)?
            }
            b'0' | b'1' | b'2' | b'3' | b'4' | b'5' | b'6' | b'7' | b'8' | b'9' => {
                self.skip()?;
                self.read_intpart(Vec::new(), c)?
            }
            b'"' => {
                self.skip()?;
                N::new(self.read_string(b'"')?)
            }
            b'|' => {
                self.skip()?;
                N::symbol(&self.read_string(b'|')?)
            }
            b';' | b'@' => {
                if read_annotations {
                    let mut annotations = self.gather_annotations(decode_embedded)?;
                    let av: N = self.demand_next_domain(read_annotations, decode_embedded)?;
                    let (existing_annotations, v) = av.pieces();
                    if let Some(vs) = existing_annotations {
                        annotations.extend_from_slice(&vs[..]);
                    }
                    N::wrap(Some(Box::new(annotations)), v)
                } else {
                    self.skip_annotations()?;
                    self.demand_next_domain(read_annotations, decode_embedded)?
                }
            }
            b':' => {
                return Err(self.syntax_error("Unexpected key/value separator between items"));
            }
            b'#' => {
                self.skip()?;
                match self.next_byte()? {
                    b'f' => N::new(false),
                    b't' => N::new(true),
                    b'{' => N::new(Set::from_iter(self.upto(b'}', read_annotations, decode_embedded)?.into_iter())),
                    b'"' => self.read_literal_binary()?,
                    b'x' => if self.next_byte()? == b'"' {
                        self.read_hex_binary()?
                    } else {
                        return Err(self.syntax_error("Expected open-quote at start of hex ByteString"));
                    },
                    b'[' => self.read_base64_binary()?,
                    b'=' => {
                        let bs_val = self.next_iovalue(true)?;
                        if let Some(anns) = bs_val.annotations() {
                            if anns.len() > 0 {
                                return Err(self.syntax_error("Annotations not permitted after #="));
                            }
                        }
                        match bs_val.value().as_bytestring() {
                            None =>
                                return Err(self.syntax_error("ByteString must follow #=")),
                            Some(bs) =>
                                crate::value::BytesBinarySource::new(bs)
                                .packed()
                                .demand_next_domain(read_annotations, decode_embedded)?
                        }
                    }
                    b'!' => Value::Embedded(decode_embedded.decode_embedded(self, read_annotations)?).wrap(),
                    other => return Err(self.syntax_error(&format!("Invalid # syntax: {:?}", other))),
                }
            }
            b'<' => {
                self.skip()?;
                let vs = self.upto(b'>', read_annotations, decode_embedded)?;
                if vs.is_empty() {
                    return Err(self.syntax_error("Missing record label"));
                }
                Value::Record(Record(vs)).wrap()
            }
            b'[' => {
                self.skip()?;
                N::new(self.upto(b']', read_annotations, decode_embedded)?)
            }
            b'{' => {
                self.skip()?;
                self.read_dictionary(read_annotations, decode_embedded)?
            }
            b'>' => return Err(self.syntax_error("Unexpected >")),
            b']' => return Err(self.syntax_error("Unexpected ]")),
            b'}' => return Err(self.syntax_error("Unexpected }")),
            other => {
                self.skip()?;
                self.read_raw_symbol(vec![other])?
            }
        }))
    }

    fn open_record(&mut self) -> ReaderResult<()> {
        self.skip_annotations()?;
        if self.peek()? != Some(b'<') { return Err(Error::Expected(ExpectedKind::Record)); }
        self.skip()?;
        Ok(())
    }

    fn open_sequence(&mut self) -> ReaderResult<()> {
        self.skip_annotations()?;
        if self.peek()? != Some(b'[') { return Err(Error::Expected(ExpectedKind::Sequence)); }
        self.skip()?;
        Ok(())
    }

    fn open_set(&mut self) -> ReaderResult<()> {
        self.skip_annotations()?;
        let mark = self.mark()?;
        match self.next_byte()? {
            b'#' => match self.next_byte()? {
                b'{' => return Ok(()),
                _ => (),
            },
            _ => (),
        }
        self.restore(&mark)?;
        Err(Error::Expected(ExpectedKind::Set))
    }

    fn open_dictionary(&mut self) -> ReaderResult<()> {
        self.skip_annotations()?;
        if self.peek()? != Some(b'{') { return Err(Error::Expected(ExpectedKind::Dictionary)); }
        self.skip()?;
        Ok(())
    }

    #[inline]
    fn boundary(&mut self, b: &B::Type) -> ReaderResult<()> {
        match b {
            B::Type {
                closing: Some(B::Item::DictionaryKey),
                opening: Some(B::Item::DictionaryValue),
            } => {
                self.skip_whitespace();
                if self.next_byte()? != b':' {
                    Err(self.syntax_error("Missing expected key/value separator"))?;
                }
            },
            _ => (),
        }
        Ok(())
    }

    fn close_compound(&mut self, b: &mut B::Type, i: &B::Item) -> ReaderResult<bool> {
        self.skip_whitespace();
        match self.peek_noeof()? {
            b'>' | b']' | b'}' => {
                self.skip()?;
                Ok(true)
            }
            _ => {
                b.shift(Some(i.clone()));
                self.boundary(b)?;
                Ok(false)
            }
        }
    }

    fn open_embedded(&mut self) -> ReaderResult<()> {
        self.skip_annotations()?;
        let mark = self.mark()?;
        match self.next_byte()? {
            b'#' => match self.next_byte()? {
                b'!' => return Ok(()),
                _ => (),
            },
            _ => (),
        }
        self.restore(&mark)?;
        Err(Error::Expected(ExpectedKind::Embedded))
    }

    fn close_embedded(&mut self) -> ReaderResult<()> {
        Ok(())
    }

    type Mark = S::Mark;

    fn mark(&mut self) -> io::Result<Self::Mark> {
        self.source.mark()
    }

    fn restore(&mut self, mark: &Self::Mark) -> io::Result<()> {
        self.source.restore(mark)
    }

    fn next_token<D: Domain, Dec: DomainDecode<D>>(
        &mut self,
        read_embedded_annotations: bool,
        decode_embedded: &mut Dec,
    ) -> io::Result<Token<D>> {
        self.skip_annotations()?;
        let mark = self.mark()?;
        Ok(match self.next_byte()? {
            b'<' => Token::Compound(CompoundClass::Record),
            b'[' => Token::Compound(CompoundClass::Sequence),
            b'{' => Token::Compound(CompoundClass::Dictionary),
            b'>' => Token::End,
            b']' => Token::End,
            b'}' => Token::End,
            b'#' => match self.next_byte()? {
                b'!' => Token::Embedded(decode_embedded.decode_embedded(self, read_embedded_annotations)?),
                b'{' => Token::Compound(CompoundClass::Set),
                _ => {
                    self.restore(&mark)?;
                    Token::Atom(self.demand_next_domain(false, decode_embedded)?)
                }
            },
            _ => {
                self.restore(&mark)?;
                Token::Atom(self.demand_next_domain(false, decode_embedded)?)
            }
        })
    }
}
